@@ -19,9 +19,11 @@ import (
 	"strings"
 	"time"
 
-	"github.com/ProtonMail/gopenpgp/v2/crypto" // MIT License
-	"github.com/pborman/getopt/v2"             // BSD-3-Clause
-	"github.com/xmppo/go-xmpp"                 // BSD-3-Clause
+	gopenpgpConst "github.com/ProtonMail/gopenpgp/v2/constants" // MIT License
+	"github.com/ProtonMail/gopenpgp/v2/crypto"                  // MIT License
+	"github.com/pborman/getopt/v2"                              // BSD-3-Clause
+	"github.com/xmppo/go-xmpp"                                  // BSD-3-Clause
+	"salsa.debian.org/mdosch/xmppsrv"                           // BSD-2-Clause
 )
 
 type configuration struct {
@@ -97,7 +99,7 @@ func main() {
 
 	// Define command line flags.
 	flagHelp := getopt.BoolLong("help", 0, "Show help.")
-	flagHTTPUpload := getopt.StringLong("http-upload", 'h', "", "Send a file via http-upload.")
+	flagHTTPUpload := getopt.ListLong("http-upload", 'h', "Send a file via http-upload. Can be invoked several times to upload multiple files.")
 	flagDebug := getopt.BoolLong("debug", 'd', "Show debugging info.")
 	flagServer := getopt.StringLong("jserver", 'j', "", "XMPP server address.")
 	flagUser := getopt.StringLong("username", 'u', "", "Username for XMPP account.")
@@ -137,6 +139,11 @@ func main() {
 	flagSSDPOff := getopt.BoolLong("ssdp-off", 0, "Disable XEP-0474: SASL SCRAM Downgrade Protection.")
 	flagSubject := getopt.StringLong("subject", 's', "", "Set message subject.")
 	flagFastOff := getopt.BoolLong("fast-off", 0, "Disable XEP-0484: Fast Authentication Streamlining Tokens.")
+	flagFastInvalidate := getopt.BoolLong("fast-invalidate", 0, "Invalidate XEP-0484: Fast Authentication Streamlining Tokens.")
+	flagPLAINAllow := getopt.BoolLong("allow-plain", 0, "Allow PLAIN authentication.")
+	flagNoRootWarning := getopt.BoolLong("suppress-root-warning", 0, "Suppress warning when run as root.")
+	flagAnon := getopt.BoolLong("anonymous", 0, "Use anonymous authentication (specify the target anon service as username.")
+	flagNoSASLUpgrade := getopt.BoolLong("no-sasl-upgrade", 0, "Disable XEP-0480: SASL Upgrade Tasks.")
 
 	// Parse command line flags.
 	getopt.Parse()
@@ -149,20 +156,23 @@ func main() {
 	case *flagVersion:
 		// If requested, show version and quit.
 		fmt.Println("Go-sendxmpp", version)
+		fmt.Println("Go-xmpp library version:", xmpp.Version)
+		fmt.Println("Xmppsrv library version:", xmppsrv.Version)
+		fmt.Println("Gopenpgp library version:", gopenpgpConst.Version)
 		system := runtime.GOOS + "/" + runtime.GOARCH
 		fmt.Println("System:", system, runtime.Version())
 		fmt.Println("License: BSD-2-clause")
 		os.Exit(0)
 		// Quit if Ox (OpenPGP for XMPP) is requested for unsupported operations like
 		// groupchat, http-upload or listening.
-	case *flagOx && *flagHTTPUpload != "":
+	case *flagOx && len(*flagHTTPUpload) != 0:
 		log.Fatal("No Ox support for http-upload available.")
 	case *flagOx && *flagChatroom:
 		log.Fatal("No Ox support for chat rooms available.")
-	case *flagHTTPUpload != "" && *flagInteractive:
+	case len(*flagHTTPUpload) != 0 && *flagInteractive:
 		log.Fatal("Interactive mode and http upload can't" +
 			" be used at the same time.")
-	case *flagHTTPUpload != "" && *flagMessageFile != "":
+	case len(*flagHTTPUpload) != 0 && *flagMessageFile != "":
 		log.Fatal("You can't send a message while using" +
 			" http upload.")
 	case *flagOx && *flagOOBFile != "":
@@ -171,10 +181,12 @@ func main() {
 		log.Fatal("No Ox support for headline messages.")
 	case *flagHeadline && *flagChatroom:
 		log.Fatal("Can't use message type headline for groupchat messages.")
+	case *flagAnon && *flagUser == "":
+		log.Fatal("Specifying a username is required when using anonymous authentication.")
 	}
 
 	// Print a warning if go-sendxmpp is run by the user root on non-windows systems.
-	if runtime.GOOS != "windows" {
+	if runtime.GOOS != "windows" && !*flagNoRootWarning {
 		// Get the current user.
 		currUser, err := osUser.Current()
 		if err != nil {
@@ -201,13 +213,14 @@ func main() {
 	// when sending raw messages to MUCs (go-sendxmpp will join the MUC automatically).
 	recipientsList := getopt.Args()
 	if (len(recipientsList) == 0 && !*flagRaw && !*flagListen && !*flagOxGenPrivKeyX25519 &&
-		!*flagOxGenPrivKeyRSA && *flagOxImportPrivKey == "") && !*flagOxDeleteNodes ||
-		(len(recipientsList) == 0 && *flagChatroom) {
+		!*flagOxGenPrivKeyRSA && *flagOxImportPrivKey == "" && !*flagFastInvalidate) &&
+		!*flagOxDeleteNodes || (len(recipientsList) == 0 && *flagChatroom) {
 		log.Fatal("No recipient specified.")
 	}
 
-	// Read configuration file if user or password is not specified.
-	if *flagUser == "" || *flagPassword == "" {
+	// Read configuration file if no user or password is specified and anonymous
+	// authentication is not used.
+	if (*flagUser == "" || *flagPassword == "") && !*flagAnon {
 		// Read configuration from file.
 		config, err := parseConfig(*flagFile)
 		if err != nil {
@@ -234,14 +247,14 @@ func main() {
 	}
 
 	// Overwrite password if specified via command line flag.
-	if *flagPassword != "" {
+	if *flagPassword != "" && !*flagAnon {
 		password = *flagPassword
 	}
 	// If no server part is specified in the username but a server is specified
 	// just assume the server is identical to the server part and hope for the
 	// best. This is for compatibility with the old perl sendxmpp config files.
 	var serverpart string
-	if !strings.Contains(user, "@") && server != "" {
+	if !strings.Contains(user, "@") && server != "" && !*flagAnon {
 		// Remove port if server contains it.
 		if strings.Contains(server, ":") {
 			serverpart, _, err = net.SplitHostPort(server)
@@ -273,16 +286,25 @@ func main() {
 
 	if !*flagFastOff {
 		fast, _ = getFastData(user, password)
-		// Reset FAST token and mechanism if expired.
-		if time.Now().After(fast.Expiry) {
+		// Reset FAST token and mechanism if expired or
+		// FastInvalidate is set.
+		if time.Now().After(fast.Expiry) || *flagFastInvalidate {
 			fast.Token = ""
 			fast.Mechanism = ""
 		}
 	}
+	if fast.Mechanism != "" && *flagSCRAMPinning != "" {
+		log.Fatalf(("FAST: %s is requested, but we have a token for %s."), *flagSCRAMPinning, fast.Mechanism)
+	}
 
 	// Use ALPN
 	var tlsConfig tls.Config
-	tlsConfig.ServerName = user[strings.Index(user, "@")+1:]
+	if !*flagAnon {
+		tlsConfig.ServerName = user[strings.Index(user, "@")+1:]
+	} else {
+		tlsConfig.ServerName = *flagUser
+		*flagSCRAMPinning = "ANONYMOUS"
+	}
 	tlsConfig.NextProtos = append(tlsConfig.NextProtos, "xmpp-client")
 	tlsConfig.InsecureSkipVerify = *flagSkipVerify
 	tlsConfig.Renegotiation = tls.RenegotiateNever
@@ -302,6 +324,10 @@ func main() {
 
 	resource := "go-sendxmpp." + getShortID()
 
+	// Check whether PLAIN authentication is disabled.
+	pinNoPLAIN, _ := parseAuthPinFile(user)
+	noPLAIN := pinNoPLAIN || !*flagPLAINAllow
+
 	// Set XMPP connection options.
 	options := xmpp.Options{
 		Host:        server,
@@ -315,17 +341,21 @@ func main() {
 		// set when NoTLS is set go-sendxmpp won't use unencrypted
 		// client-to-server connections.
 		// See https://pkg.go.dev/github.com/xmppo/go-xmpp#Options
-		NoTLS:         !*flagDirectTLS,
-		StartTLS:      !*flagDirectTLS,
-		Debug:         *flagDebug,
-		TLSConfig:     &tlsConfig,
-		Mechanism:     *flagSCRAMPinning,
-		SSDP:          !*flagSSDPOff,
-		UserAgentSW:   resource,
-		UserAgentID:   clientID,
-		Fast:          !*flagFastOff,
-		FastToken:     fast.Token,
-		FastMechanism: fast.Mechanism,
+		NoTLS:          !*flagDirectTLS,
+		StartTLS:       !*flagDirectTLS,
+		Debug:          *flagDebug,
+		DebugWriter:    os.Stdout,
+		TLSConfig:      &tlsConfig,
+		Mechanism:      *flagSCRAMPinning,
+		SSDP:           !*flagSSDPOff,
+		UserAgentSW:    resource,
+		UserAgentID:    clientID,
+		Fast:           !*flagFastOff,
+		FastToken:      fast.Token,
+		FastMechanism:  fast.Mechanism,
+		FastInvalidate: *flagFastInvalidate,
+		NoPLAIN:        noPLAIN,
+		NoSASLUpgrade:  *flagNoSASLUpgrade,
 	}
 
 	// Read message from file.
@@ -338,10 +368,10 @@ func main() {
 
 	// Skip reading message if '-i' or '--interactive' is set to work with e.g. 'tail -f'.
 	// Also for listening mode and Ox key handling.
-	if !*flagInteractive && !*flagListen && *flagHTTPUpload == "" &&
+	if !*flagInteractive && !*flagListen && len(*flagHTTPUpload) == 0 &&
 		!*flagOxDeleteNodes && *flagOxImportPrivKey == "" &&
 		!*flagOxGenPrivKeyX25519 && !*flagOxGenPrivKeyRSA && *flagOOBFile == "" &&
-		message == "" {
+		!*flagFastInvalidate && message == "" {
 		scanner := bufio.NewScanner(os.Stdin)
 		for scanner.Scan() {
 			if message == "" {
@@ -363,28 +393,16 @@ func main() {
 	// Exit if message is empty.
 	if message == "" && !*flagInteractive && !*flagListen && !*flagOxGenPrivKeyRSA &&
 		!*flagOxGenPrivKeyX25519 && *flagOxImportPrivKey == "" &&
-		!*flagOxDeleteNodes && *flagHTTPUpload == "" && *flagOOBFile == "" {
+		!*flagOxDeleteNodes && len(*flagHTTPUpload) == 0 && *flagOOBFile == "" &&
+		!*flagFastInvalidate {
 		os.Exit(0)
 	}
 
 	// Connect to server.
-	client, err := connect(options, *flagDirectTLS)
+	client, err := connect(options, *flagDirectTLS, *flagAnon)
 	if err != nil {
 		if fast.Token != "" {
-			// Reset FAST token and mechanism if FAST login failed.
-			fast.Token = ""
-			fast.Mechanism = ""
-			fast.Expiry = time.Now()
-			err := writeFastData(user, password, fast)
-			if err != nil {
-				fmt.Println(err)
-			}
-			options.FastToken = ""
-			// Try to connect to server without FAST.
-			client, err = connect(options, *flagDirectTLS)
-			if err != nil {
-				log.Fatal(err)
-			}
+			log.Fatal(fmt.Errorf("could not connect using FAST authentication: %v", err))
 		} else {
 			log.Fatal(err)
 		}
@@ -401,6 +419,22 @@ func main() {
 			fmt.Println(err)
 		}
 	}
+	// Delete stored fast data when FastInvalidate is set.
+	if *flagFastInvalidate && *flagFastOff {
+		err := deleteFastData(user)
+		if err != nil {
+			fmt.Println(err)
+		}
+	}
+	// If authentication is not yet pinned to not use PLAIN and a SCRAM mechanism is
+	// used, write the auth pin file.
+	if !pinNoPLAIN && (strings.HasPrefix(client.Mechanism, "SCRAM") ||
+		strings.HasPrefix(client.Mechanism, "HT")) {
+		err = writeAuthPinFile(user)
+		if err != nil {
+			log.Println("could not write authentication mechanism pin file:", err)
+		}
+	}
 	iqc := make(chan xmpp.IQ, defaultBufferSize)
 	msgc := make(chan xmpp.Chat, defaultBufferSize)
 	ctx, cancel := context.WithCancel(context.Background())
@@ -492,9 +526,10 @@ func main() {
 		}
 	}
 
-	if *flagHTTPUpload != "" {
-		message, err = httpUpload(client, iqc, tlsConfig.ServerName,
-			*flagHTTPUpload, timeout)
+	var uploadMessages []string
+	if len(*flagHTTPUpload) != 0 {
+		uploadMessages, err = httpUpload(client, iqc,
+			tlsConfig.ServerName, *flagHTTPUpload, timeout)
 		if err != nil {
 			cancel()
 			closeAndExit(client, err)
@@ -697,19 +732,21 @@ func main() {
 		}
 	default:
 		for _, recipient := range recipients {
-			if message == "" {
+			if message == "" && len(uploadMessages) == 0 {
 				break
 			}
 			switch {
-			case *flagHTTPUpload != "":
-				_, err = client.Send(xmpp.Chat{
-					Remote: recipient.Jid,
-					Type:   msgType, Ooburl: message, Text: message,
-					Subject: *flagSubject,
-				})
-				if err != nil {
-					fmt.Println("Couldn't send message to",
-						recipient.Jid)
+			case len(*flagHTTPUpload) != 0:
+				for _, message = range uploadMessages {
+					_, err = client.Send(xmpp.Chat{
+						Remote: recipient.Jid,
+						Type:   msgType, Ooburl: message, Text: message,
+						Subject: *flagSubject,
+					})
+					if err != nil {
+						fmt.Println("Couldn't send message to",
+							recipient.Jid)
+					}
 				}
 			// (Hopefully) temporary workaround due to go-xmpp choking on URL encoding.
 			// Once this is fixed in the lib the http-upload case above can be reused.
