@@ -7,6 +7,7 @@ package main
 import (
 	"encoding/base64"
 	"fmt"
+	"io"
 	"log"
 	"os"
 	"runtime"
@@ -315,19 +316,19 @@ func oxPublishPubKey(jid string, client *xmpp.Client, iqc chan xmpp.IQ, pubKey *
 }
 
 func oxGetPrivKeyLoc(jid string) (string, error) {
-	dataDir, err := getDataPath(strings.Replace(strings.Replace(jid, "@", "_at_", -1), ".", "_", -1))
+	dataDir, err := getDataPath(fsFriendlyJid(jid), true)
 	if err != nil {
 		return strError, fmt.Errorf("oxGetPrivKeyLoc: %w", err)
 	}
-	oldDataDir, err := getDataPath("oxprivkeys/")
+	oldDataDir, err := getDataPath("oxprivkeys/", false)
 	if err != nil {
 		return strError, fmt.Errorf("oxGetPrivKeyLoc: %w", err)
 	}
 	// TODO: Remove handling of oldDataFile in a later version when it's very likely that there are no
 	// more versions in use using the oldDataFile (<0.8.3).
 	oldDataFile := oldDataDir + base64.StdEncoding.EncodeToString([]byte(jid))
-	oldDataFile2 := oldDataDir + strings.Replace(jid, "@", "_at_", -1)
-	oldDataFile3 := oldDataDir + strings.Replace(strings.Replace(jid, "@", "_at_", -1), ".", "_", -1)
+	oldDataFile2 := oldDataDir + strings.ReplaceAll(jid, "@", "_at_")
+	oldDataFile3 := oldDataDir + fsFriendlyJid(jid)
 	dataFile := dataDir + "oxprivkey"
 	if _, err := os.Stat(oldDataFile); err == nil {
 		err := os.Rename(oldDataFile, dataFile)
@@ -347,11 +348,22 @@ func oxGetPrivKeyLoc(jid string) (string, error) {
 			return dataFile, err
 		}
 	}
+	dir, err := os.Open(oldDataDir)
+	if err == nil {
+		_, err = dir.ReadDir(1)
+		// Delete old data dir if empty.
+		if err == io.EOF {
+			err = os.Remove(oldDataDir)
+			if err != nil {
+				fmt.Printf("oxGetPrivKeyLoc: failed to delete old private key location: %v\n", err)
+			}
+		}
+	}
 	return dataFile, nil
 }
 
 func oxGetPubKeyLoc(fingerprint string) (string, error) {
-	dataDir, err := getDataPath("oxpubkeys/")
+	dataDir, err := getDataPath("oxpubkeys/", true)
 	if err != nil {
 		return strError, fmt.Errorf("oxGetPubKeyLoc: %w", err)
 	}
